/*
 *
 *  Low level packet transmission code for Unix sockets.
 *
 *
 *  There are 8 routines exported from this file:
 *
 *   Net_Initialize (service_id)
 *       This routine must be invoked prior to any other calls.
 *       The service_id provided by the user defines the inet "port"
 *       number to be used by this process as its service address.
 *       This number must be unique on the machine in which the 
 *       program executes and must be greater than 5000 and less than 64K.
 *       You should pick a random number in this range to avoid conflicts
 *       with other users/servers. A return code of NET_SERVER_ID_INUSE
 *       indicates the requested port number is already in use. You should
 *       retry using a different value. 
 *
 *       In order to provide a symmetric, full-duplex connection between
 *       clients and servers, each process should be assigned a service_id.
 *       The client's service id is necessary to allow the server to 
 *       reestablish a connection to the client after a client failure.
 *
 * 
 *   Net_Open (remote_host_name, remote_service_id, rtnd_connection_id)
 *       This routine must be invoked to initiate a session with a remote
 *       program. The remote machine must have already started and invoked
 *       Net_Initialize. The "remote_host_name" is the machine name on which 
 *       the program executes. The "remote_service_id" is the port number 
 *       selected by the remote program when it invoked Net_Initialize. 
 *       This routine establishes a network session with the remote program 
 *       and returns a "connection id" to the caller to be used in subsequent 
 *       calls to Net_Send. 
 * 
 *       Note that connection id uniquely identifies the remote service. 
 *       The connection id does not change if the remote process fails 
 *       and is restarted. 
 *
 *
 *   Net_Send (connection_id, ptr_to_data_buffer, data_length)
 *       This routine sends the packet to the remote machine specified
 *       by the connection id. The packet data is not interpreted, but
 *       it cannot be longer than NET_MAX_PACKET_SIZE.
 *
 *       Attempting to send data to a failed process MAY result in this 
 *       routine returning NET_PROCESS_FAILED. A subsequent call to 
 *       Net_Send will attempt to reestablish the connection with the 
 *       restarted process offering the same remote_service_id.
 *
 *      Note that this routine, like Unix, does not guarantee that the
 *      remote host has received the data before the call returns.
 *      Consequently, sending data to a recently failed process MAY result 
 *      in a NET_SUCCESS return code. Eventually the underlying network
 *      protocol will decide the remote process has failed and a subsequent
 *      call to Net_Send or Net_Receive will return NET_PROCESS_FAILED. 
 *
 *
 *   Net_Receive (ptr_to_data_buffer, rtnd_data_len, rtnd_connection_id)
 *       This routine blocks the caller, waiting for a message to be
 *       received from any connection that the program has established.
 *       The data buffer provided by the caller must be at least
 *       NET_MAX_PACKET_SIZE bytes long. Once a packet is received,
 *       this routine returns the packet data in the provided buffer,
 *       the length of the data is returned in "rtnd_data_len" and
 *       the sender's connection id is returned in "rtnd_connection_id".
 *
 *       Note that this routine also accepts new connections to the
 *       service offered by this process (the service_id was defined
 *       when the program called Net_Initialize). The caller is not
 *       informed when the new connection is established, but will 
 *       returned a new "connection_id" when the first packet is 
 *       received on that connection. 
 *
 *       In the event of a remote process failure, this routine will 
 *       return NET_PROCESS_FAILED with the rtnd_connection_id set 
 *       to the connection id of the failed process. Subsequent calls
 *       to Net_Receive will no longer wait for messages from the
 *       failed process until the connection is reestablished.
 *
 *
 *   Net_Poll (timeout, ptr_to_data_buffer, rtnd_data_len, rtnd_connection_id)
 *       This routine waits for a message from any connection that the
 *       program has established, but the calling program is blocked
 *       only for the "timeout" seconds. If no message is received 
 *       within the specifed timeout, this return will return a value of
 *       NET_TIMEOUT. The timeout value is in seconds, a value of 0
 *       returns immediately. See Net_Receive for a description of the
 *       remaining parameters.
 *       
 * 
 *   Net_Trace ()
 *       This routine toggles the trace flag and returns the value assigned
 *       to the flag. When tracing is enabled it shows all packets sent and 
 *       received, as well as sessions being started and ended.
 *
 *
 *   Net_Pause ()
 *       This routine toggles the pause flag and returns the value assigned
 *       to the flag. When pausing is enabled, the user is prompted for 
 *       input via stdin before messages are sent or received.
 *
 *
 *   Net_Debug ()
 *       This routine prints the connection table maintained by the low-level
 *       code. It is used solely for debugging.
 *
 *
 * All routines (except Net_Trace and Net_Debug) return a status code as the 
 * function's value. The return codes are defined in net.h as follows:
 *   NET_SUCCESS            0     -- ok
 *   NET_SERVER_ID_INUSE    1     -- requested server id is not unique
 *   NET_PROCESS_FAILED     2     -- remote process failed
 *   NET_TIMEOUT            3     -- receive request timed out
 *   NET_IO_ERROR          -1     -- error occurred during i/o operation
 *   NET_UNIX_ERROR        -2     -- error returned from Unix operation
 *   NET_USER_ERROR        -3     -- application program error
 *
 */


/*
 * Define constants 
 */
#define NET_MAX_PACKET_SIZE 20480    /* Max data allowed to send at once */
#define NET_MAX_CONNECTIONS   256    /* Max # of concurrent net sessions */
#define NET_MIN_SERVER_ID    5000    /* = IPPORT_USERRESERVED */
#define NET_MAX_SERVER_ID   65535    /* must fit in a u_short */


/*
 * Define return codes
 */
#define NET_SUCCESS            0     /* ok */
#define NET_SERVER_ID_INUSE    1     /* requested server id is not unique */
#define NET_PROCESS_FAILED     2     /* remote process failed */
#define NET_TIMEOUT            3     /* receive request timed out */
#define NET_IO_ERROR          -1     /* error during an i/o operation */
#define NET_UNIX_ERROR        -2     /* error returned from Unix operation */
#define NET_USER_ERROR        -3     /* application program error */



/* Allow C++ to include this .h file */
#ifdef __cplusplus
#  define NET_EXTERN extern "C"
#else
#  define NET_EXTERN extern
#endif  __cplusplus



/*
 * Define prototypes for each routine.
 */

extern int Net_Initialize( int  server_id );

NET_EXTERN int Net_Open
   (
    char *remote_server_host,	/* Host name for remote process */
    int   remote_server_id,     /* Unique Id for server at that host */
    int  *rtnd_conn_id          /* returned connection id */
   );

NET_EXTERN int Net_Send
   (
    int   conn_id,		/* connection id to send packet to */
    void *data_ptr,		/* ptr to data buffer to send */
    int   data_len		/* length of data to send */
   );

NET_EXTERN int Net_Receive
   (
    void *buffer_ptr,		/* ptr to buffer for incoming packet 
				 * MUST be at least MAX_PACKET_SIZE long */
    int  *rtnd_data_len,	/* returned length of packet received */
    int  *rtnd_conn_id		/* returned connection id of sender */
   );

NET_EXTERN int Net_Poll
   (
    int   timeout,              /* timeout value in seconds */
    void *buffer_ptr,		/* ptr to buffer for incoming packet 
				 * MUST be at least MAX_PACKET_SIZE long */
    int  *rtnd_data_len,	/* returned length of packet received */
    int  *rtnd_conn_id		/* returned connection id of sender */
   );

NET_EXTERN int Net_Trace	/* Toggle trace flag for network code 
				 * Returns current state of trace flag. */
  (
   void
  );

NET_EXTERN int Net_Pause	/* Toggle pause flag for network code 
				 * Returns current state of pause flag. */
  (
   void
  );

NET_EXTERN void Net_Debug	/* Print connection table for debugging */
  (
   void
  );

















