/*
 *
 *  Low level packet transmission code for Unix sockets.
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>

#include "socklink.h"

extern FILE *dbgPtr;

#define VERBOSE  1	/* set to 1 to print messages for process failures */

/*
 * Define constants
 */
#define MAX_HOST_NAME 64
#define NULL_SOCKET_ID -1
#define NULL_CONNECTION_ID -1
#ifndef FALSE
#  define FALSE 0
#endif
#ifndef TRUE
#  define TRUE 1
#endif


/* 
 * Define types
 */
typedef struct {
  int     host_socket_id;
  u_short host_port_id;
  u_long  host_ip_addr;
  int     host_failed;
  char    host_name[MAX_HOST_NAME];
} host_table_t;


/*
 * Define static variables
 */
static int     net_trace_flag = 0;
static int     net_pause_flag = 0;
static int     net_initialized = 0;
static int     net_socket_id = 0;
static u_short net_server_id = 0;  /* actually an inet port id */

static fd_set  net_fd_set;
static fd_set  net_fd_read_set;
static int     net_fd_read_pending;

static int          net_host_entries = 0;
static host_table_t net_hosts[NET_MAX_CONNECTIONS];



/*
 * Define Macros 
 */
#if !defined(MIN)
#  define MIN(a,b) ((a) <= (b) ? a : b)
#endif

#if VERBOSE > 0
#  define PRINTF  printf
#  define FPRINTF fprintf
#  define PERROR  perror
#else
#  define PRINTF  (void)
#  define FPRINTF (void)
#  define PERROR  (void)
#endif


static void Net_Print_Buffer
  (
   void *voidP,
   int   dataLen
  )
{
  int i;
  char *dataP = (char *) voidP;
  
  while (dataLen >= 16) {
    printf ("%08x:  ", dataP);
    for (i = 0; i < 16; i++) {
      printf ("%02x", dataP[i]);
      if ((i % 4) == 3) printf (" ");
    }
    printf (" *");
    for (i = 0; i < 16; i++) {
      printf ("%c", (isprint(dataP[i])) ? dataP[i] : '.');
    }
    printf ("*\n");
    dataP += 16;
    dataLen -= 16;
  }
  if (dataLen > 0) {
    printf ("%08x:  ", dataP);
    for (i = 0; i < dataLen; i++) {
      printf ("%02x", dataP[i]);
      if ((i % 4) == 3) PRINTF (" ");
    }
    for ( ; i < 16; i++) {
      printf ("  ");
      if ((i % 4) == 3) printf (" ");
    }
    printf (" *");
    for (i = 0; i < dataLen; i++) {
      printf ("%c", (isprint(dataP[i])) ? dataP[i] : '.');
    }
    for ( ; i < 16; i++) {
      printf (" ");
    }
    printf ("*\n");
  }

  return;
}


void Net_Disconnect(int host)
{
  int s;
  host_table_t *hostP;
  struct sockaddr_in name;

  hostP = &net_hosts[host-1];
  if (hostP->host_socket_id == NULL_SOCKET_ID) 
    return;

  if (net_trace_flag) {
    u_long addr = ntohl(hostP->host_ip_addr);
    printf ("Net_Close: Disconnecting from "
	    "conn-id=%d, host=\"%s\" addr=[%d.%d.%d.%d] port=%d\n",
	   host, hostP->host_name, 
	   (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	   hostP->host_port_id);
  }

  FD_CLR(hostP->host_socket_id, &net_fd_set);
  (void) close (hostP->host_socket_id);
  hostP->host_socket_id = NULL_SOCKET_ID;

  return;
}



static int Net_AddHost 
  (
   int     sockId,
   u_short portId,
   u_long  ipAddr,
   char   *name,
   int    *rtndConnId
  )
{
  int hostId;
  host_table_t *hostP;

  /* Check if host already in table */
  for (hostP = net_hosts; hostP < &net_hosts[net_host_entries]; hostP++) {
    if ((hostP->host_ip_addr == ipAddr) &&
	(hostP->host_port_id == portId))
      goto found;
  }

  /* allocate space in table for new host */
  if (net_host_entries >= NET_MAX_CONNECTIONS) {
    fprintf (stderr, "Net_Open: Too many open connections, max of %d\n",
	    NET_MAX_CONNECTIONS);
    return (NET_USER_ERROR);
  }
  
  hostP = &net_hosts[ net_host_entries ];
  hostP->host_socket_id = NULL_SOCKET_ID;
  hostP->host_failed = FALSE;
  net_host_entries++;

found:

  hostId = (hostP - net_hosts) + 1;

  if (hostP->host_socket_id != NULL_SOCKET_ID) {
    Net_Disconnect(hostId);
  }

  hostP->host_socket_id = sockId;
  hostP->host_port_id =  portId;
  hostP->host_ip_addr = ipAddr;
  strncpy (hostP->host_name,	         /* to */
	   name,                         /* from */
	   sizeof(hostP->host_name));    /* length */

  *rtndConnId = hostId;
  return (NET_SUCCESS);
}


static int Net_Accept 
  (
   int     reportFailures,	/* =TRUE to report process failures */
   int    *rtndConnId
  )
{
  int    rc;
  int    s, len;
  int    namelen;
  u_long addr;
  u_long nlen;
  u_short remote_port_id;
  struct sockaddr_in  name;
  host_table_t *hostP;

  *rtndConnId = NULL_CONNECTION_ID;

  namelen = sizeof (name);
  s = accept(net_socket_id, (struct sockaddr *)&name, &namelen);
  if (s == -1) {
    perror("accept");
    return(NET_UNIX_ERROR);
  }
	
  if (net_trace_flag) {
    addr = ntohl(name.sin_addr.s_addr);
    printf ("Net_Recv: Accepting new session from [%d.%d.%d.%d]\n",
	   (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff);
  }

  /* wait for initial message to provide remote service id */
  len = read(s, (char *) &remote_port_id, sizeof(remote_port_id));
  if (len != sizeof(remote_port_id)) {
    perror ("read");
    addr = ntohl(name.sin_addr.s_addr);
    printf ("Net_Recv: Failed starting new session with [%d.%d.%d.%d]\n",
	   (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff);
    return (NET_UNIX_ERROR);
  }
	
  remote_port_id = ntohs(remote_port_id);
  
  /* 
   * Locate host in table
   * to see if it has failed and restarted
   */
  for (hostP = net_hosts; 
       hostP < &net_hosts[net_host_entries]; hostP++) {
    if ((hostP->host_ip_addr == name.sin_addr.s_addr) &&
	(hostP->host_port_id == remote_port_id)) {
      if (hostP->host_socket_id != NULL_SOCKET_ID) 
	hostP->host_failed = TRUE;
      break;
    }
  }

  rc = Net_AddHost (s, remote_port_id, name.sin_addr.s_addr, "*REMOTE*", 
		    rtndConnId);
  if (rc != NET_SUCCESS) {
    close (s);
    return(rc);
  }
  
  FD_SET(s, &net_fd_set);

  if ((reportFailures) && (hostP->host_failed)) {
    hostP->host_failed = FALSE;	/* only report failure once */
    return(NET_PROCESS_FAILED);
  }

  return (NET_SUCCESS);
}	  

  

static int Net_Reconnect(int host)
{
  int s, rc;
  int connId;
  host_table_t *hostP;
  struct sockaddr_in name;
  u_short my_port_id;
  u_long addr;
  struct timeval timeout;
  int nfds;
  fd_set fd;
      
      

  hostP = &net_hosts[host-1];

  if (hostP->host_socket_id != NULL_SOCKET_ID) {
    fprintf (stderr, "Net_Open: Internal error connecting "
	     "to host=\"%s\" addr=[%d,%d,%d,%d] port=%d\n",
	    hostP->host_name, 
	     (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	    hostP->host_port_id);
    return (NET_UNIX_ERROR);   /* Internal error -- blame it on unix */
  }

  /*
   * Hack! Before connecting to remote process, check if remote process
   * is trying to connect to us. This minimizes the window in which
   * concurrent connect's cause the two sides to be out of synch. 
   * Note that the problem can still occur and will cause sporadic
   * "failures" until the two sides synchronize.
   */
  if (net_socket_id > 0) { 
    while (1) {
      FD_ZERO(&fd);
      FD_SET(net_socket_id, &fd);
      timeout.tv_sec = 0;
      timeout.tv_usec = 0;
      
      nfds = select(FD_SETSIZE, &fd, 0L, 0L, &timeout);
      if (nfds < 0) {
	perror("select");
	return (NET_UNIX_ERROR);
      }
      if (nfds == 0) break; /* timeout -- no new connects */
      
      rc = Net_Accept (FALSE, &connId);	/* don't report process failures */
      if (rc != NET_SUCCESS) 
	return(rc);
    }
    
    if (hostP->host_socket_id != NULL_SOCKET_ID)
      return (NET_SUCCESS);
  }

  if ((s = socket(AF_INET,SOCK_STREAM,0)) == -1) {
    perror("socket");
    return (NET_UNIX_ERROR);
  }

  bzero((char *)&name, sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(hostP->host_port_id);
  bcopy((char *)&hostP->host_ip_addr, 
	(char *)&name.sin_addr, sizeof(name.sin_addr));

  if (net_trace_flag) {
    addr = ntohl(name.sin_addr.s_addr);
    printf ("Net_Open: Attempting to connect to [%d.%d.%d.%d]\n",
	   (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff);
  }

  if (connect(s, (struct sockaddr *) &name, sizeof(name)) == -1) {
    PERROR("connect");
    addr = ntohl(name.sin_addr.s_addr);
    FPRINTF (stderr, "Net_Open: Failed to connect "
	     "to host=\"%s\" addr=[%d,%d,%d,%d] port=%d\n", 
	    hostP->host_name, 
	     (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	    hostP->host_port_id);
    return (NET_USER_ERROR);
  }

  /* Send my callback port id to make the connection symmetric */
  my_port_id = htons(net_server_id);
  if (write(s, (char *) 
	    &my_port_id, sizeof(my_port_id)) != sizeof(my_port_id)) {
    perror("write");
    addr = ntohl(name.sin_addr.s_addr);
    fprintf (stderr, 
	     "Net_Open: Failed initializing "
	     "host=\"%s\" addr=[%d,%d,%d,%d] port=%d\n", 
	     hostP->host_name, 
	     (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	     hostP->host_port_id);
    close(s);
    return (NET_UNIX_ERROR);
  }

  hostP->host_socket_id = s;
  FD_SET (s, &net_fd_set);

  if (net_trace_flag) {
    u_long addr = ntohl(name.sin_addr.s_addr);
    printf ("Net_Open: Connected to "
	    "conn-id=%d, host=\"%s\" addr=[%d.%d.%d.%d] port=%d\n",
	    host, hostP->host_name, 
	    (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	    hostP->host_port_id);
  }
    
  return (NET_SUCCESS);
}


int Net_Initialize
  (
   int  server_id		/* Unique Id for each server
    				 * =0 if process is a client */
  )
{
  int			s;
  int                   rc;
  int                   set;
  int                   connId;
  u_short               portId;
  struct sockaddr_in	name;
  

  if (net_initialized) {
    fprintf (stderr, "Net_Initialize: already initialized\n");
    return(NET_USER_ERROR);
  }

  FD_ZERO(&net_fd_set);
  FD_ZERO(&net_fd_read_set);
  net_fd_read_pending = 0;

  if (server_id != 0) {

    /*
     * Server initialization 
     */
    if ((s = socket(AF_INET,SOCK_STREAM,0)) == -1) {
      perror("socket");
      return (NET_UNIX_ERROR);
    }

    set = 1;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, 
		   (char *)&set, sizeof(set)) != 0) {
      perror("setsockopt");
      close (s);
      return (NET_UNIX_ERROR);
    }
    set = 32768;
    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, 
		   (char *)&set, sizeof(set)) != 0) {
      perror("setsockopt");
      close (s);
      return (NET_UNIX_ERROR);
    }
    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, 
		   (char *)&set, sizeof(set)) != 0) {
      perror("setsockopt");
      close (s);
      return (NET_UNIX_ERROR);
    }

    bzero((char *)&name,sizeof(name));
    name.sin_family = AF_INET;
    name.sin_addr.s_addr = INADDR_ANY;
    portId = (u_short) server_id;
    name.sin_port = htons(portId);

    if (bind(s, (struct sockaddr *)&name, sizeof(name)) == -1) {
      perror("bind");
      close (s);
      return (NET_SERVER_ID_INUSE);
    }

    /* backlogs?  Allow a backlog of 5 */
    if (listen(s,5) == -1) {
      perror("listen");
      close (s);
      return (NET_UNIX_ERROR);
    }

    rc = Net_AddHost (s, portId, name.sin_addr.s_addr, "*LOCAL*", &connId);
    if (rc != NET_SUCCESS) {
      close (s);
      return (rc);
    }

    net_socket_id = s;
    FD_SET (s, &net_fd_set);
  }

  net_initialized = 1;
  net_server_id = (u_short) server_id;

  return (NET_SUCCESS);
}



int Net_Open
   (
    char *remote_server_host,	/* Host name for remote process */
    int   remote_server_id,     /* Unique Id for server at that host */
    int  *rtnd_conn_id          /* returned connection id */
   )
{
  int                 rc;
  int                 connId;
  u_short             portId;
  struct hostent     *hentP;
  struct sockaddr_in  name;

  *rtnd_conn_id = NULL_CONNECTION_ID;

  if (!net_initialized) {
    fprintf (stderr, "Net_Open: not initialized\n");
    return(NET_USER_ERROR);
  }
  
  hentP = gethostbyname(remote_server_host);
  if (!hentP) {
    fprintf (stderr, 
	     "Net_Open: No host entry found for \"%s\"\n", 
	     remote_server_host);
    return(NET_USER_ERROR);
  }

  portId = (u_short) remote_server_id;
  bzero ((char *)&name, sizeof(name));
  bcopy((char *) hentP->h_addr_list[0],
	(char *) &name.sin_addr, 
	sizeof(name.sin_addr));


  rc = Net_AddHost (NULL_SOCKET_ID, portId,
		    name.sin_addr.s_addr, 
		    remote_server_host, &connId);
  if (rc != NET_SUCCESS) return (rc);

  rc = Net_Reconnect (connId);
  if (rc == NET_SUCCESS) 
    *rtnd_conn_id = connId;

  return (rc);
}



int Net_Send
   (
    int   conn_id,		/* connection id to send packet to */
    void *data_ptr,		/* ptr to data buffer to send */
    int   data_len		/* length of data to send */
   )
{
  int     s;
  int     rc;
  u_long  nlen;
  u_long  addr;
  int	  towrite, written;
  int     so_error, so_error_len;
  char   *dataP;
  host_table_t *hostP;


  if ((conn_id <= 0) || (conn_id > net_host_entries)) {
    fprintf (stderr, "Net_Send: Invalid connection id %d\n", conn_id);
    return (NET_USER_ERROR);
  }
  
  if (data_len > NET_MAX_PACKET_SIZE) {
    fprintf (stderr, "Net_Send: Packet too large, size=%d max size=%d\n", 
	    data_len, NET_MAX_PACKET_SIZE);
    return (NET_USER_ERROR);
  }

  /* Convert user's connection id into host table slot number */
  hostP = &net_hosts[ conn_id-1 ];
  s = hostP->host_socket_id;

  /* Check if still connected to remote host */
  if (s == NULL_SOCKET_ID) {
    rc = Net_Reconnect (conn_id);
    if (rc != NET_SUCCESS) return (rc);
    s = hostP->host_socket_id;
  }

  /* Check if remote host failed & restarted */
  if (hostP->host_failed) {
    hostP->host_failed = FALSE;
    return (NET_PROCESS_FAILED);
  }


  if (net_trace_flag) {
    addr = hostP->host_ip_addr;
    printf ("Net_Send: Sending %d bytes to "
	    "conn-id %d, host=\"%s\" [%d,%d,%d,%d] port=%d\n",
	   data_len, conn_id, hostP->host_name, 
	   (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	   hostP->host_port_id);
    Net_Print_Buffer (data_ptr, data_len);
  }

  if (net_pause_flag) {
    char user[256];
    addr = hostP->host_ip_addr;
    printf ("Net_Send: Press <ENTER> to send data.  Enter 'd' to discard.\n");
    gets(user);
    if ((strlen(user) > 0) && ((user[0] == 'd') || (user[0] == 'D')))
      return (NET_SUCCESS);
  }

#if 0
  /* Check if remote host is still available .
   * This is disgusting, but the return code from Unix write
   * only indicates the data was successfully copied into
   * the kernel buffer and does not indicate the status of the 
   * connection or the remote machine.
   */
  { int nfds;
    u_long rlen, rnlen;
    fd_set fd;
    struct timeval timeout;
    
    FD_ZERO(&fd);
    FD_SET(s, &fd);
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    nfds = select(FD_SETSIZE, &fd, 0L, 0L, &timeout);
    if (nfds < 0) {
      perror("select");
      return (NET_UNIX_ERROR);
    }
    if (nfds > 0) {
      rlen = recv(s, (char *) &rnlen, sizeof(rnlen), MSG_PEEK);
      if (rlen != sizeof(rnlen)) {
	PERROR ("recv");
	addr = ntohl(hostP->host_ip_addr);
	FPRINTF (stderr, 
		 "Net_Send: Error sending data to "
		 "host=\"%s\" [%d,%d,%d,%d] port=%d\n",
		 hostP->host_name, 
		 (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
		 hostP->host_port_id);
	Net_Disconnect (conn_id);
	hostP->host_failed = FALSE;	/* only report failure once */
	return (NET_PROCESS_FAILED);
      }
    }
  }
#endif

  nlen = data_len;
  nlen = htonl(nlen);
  if (write(s, (char *) &nlen, sizeof(nlen)) != sizeof(nlen)) {
    PERROR("write (data size)");
    addr = ntohl(hostP->host_ip_addr);
    FPRINTF (stderr, "Net_Send: Error sending data "
	     "to host=\"%s\" [%d,%d,%d,%d] port=%d\n",
	     hostP->host_name, 
	     (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	   hostP->host_port_id);
    Net_Disconnect(conn_id);
    hostP->host_failed = FALSE;	/* only report failure once */
    return (NET_PROCESS_FAILED);
  }

  towrite = data_len;
  dataP = (char *) data_ptr;
  while (towrite) {
    if ((written = write(s, dataP, towrite)) == -1) {
      PERROR("write (data)");
      addr = ntohl(hostP->host_ip_addr);
      FPRINTF (stderr, "Net_Send: Error sending data "
	       "to host=\"%s\" [%d,%d,%d,%d] port=%d\n",
	       hostP->host_name, 
	       (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	       hostP->host_port_id);
      Net_Disconnect(conn_id);
      hostP->host_failed = FALSE;	/* only report failure once */
      return (NET_PROCESS_FAILED);
    }

    towrite -= written;
    dataP += written;
  }
  
  return (NET_SUCCESS);
}



int Net_Poll
   (
    int   timeout_val,          /* timeout value in seconds, -1 = no timeout */
    void *buffer_ptr,		/* ptr to buffer for incoming packet 
				 * MUST be at least MAX_PACKET_SIZE long */
    int  *rtnd_data_len,	/* returned length of packet received */
    int  *rtnd_conn_id		/* returned connection id of sender */
   )
{
  int    rc;
  int    host, s, len;
  int    nfds;
  int    connId;
  int    namelen;
  u_long nlen;
  u_short remote_port_id;
  u_long addr;
  int    nhosts;
  int	 toread, hasread;
  int    dataLen;
  char  *dataP;
  struct sockaddr_in  name;
  host_table_t *hostP;
  struct timeval timeout, *timeoutP;


  /* 
   * Check for failed host that restarted since the last call to Net_Receive
   */
  for (host = 0; host < net_host_entries; host++) {
    hostP = &net_hosts[host];
    if (hostP->host_failed) {
      *rtnd_conn_id = host+1;
      *rtnd_data_len = 0;
      hostP->host_failed = FALSE;
      return (NET_PROCESS_FAILED);
    }
  }
  

CallSelect:

  /* Check if there is a read pending from our previous call */
  if (net_fd_read_pending == 0) {

    /* copy combined fd_set, so that original is not modified */
    net_fd_read_set = net_fd_set;

    if (timeout_val >= 0) {
      timeout.tv_sec = timeout_val;
      timeout.tv_usec = 0;
      timeoutP = &timeout;
    } else
      timeoutP = (struct timeval *) 0;

    nfds = select(FD_SETSIZE, &net_fd_read_set, 0L, 0L, timeoutP);
    if (nfds < 0) {
      perror("select");
      return (NET_UNIX_ERROR);
    }
    if (nfds == 0) { /* timeout */
      *rtnd_data_len = 0;
      *rtnd_conn_id = 0;
      return (NET_TIMEOUT);
    }

    /* Check for new connections to servers */
    if (net_socket_id > 0) {
      if (FD_ISSET(net_socket_id, &net_fd_read_set)) {
	
	FD_CLR(net_socket_id, &net_fd_read_set);

	rc = Net_Accept(TRUE, rtnd_conn_id); /* report process failures */
	if (rc != NET_SUCCESS) {
	  *rtnd_data_len = 0;
	  return(rc);
	}
      }
    }
  }
  
  /* Check for incoming messages */
  for (host = net_fd_read_pending; host < net_host_entries; host++) {

    hostP = &net_hosts[host];
    s = hostP->host_socket_id;
    if (s == NULL_SOCKET_ID) continue;

    if (FD_ISSET (s, &net_fd_read_set)) {
      
      FD_CLR(s, &net_fd_read_set);
      net_fd_read_pending = host+1;

      /* Return data to caller */
      *rtnd_conn_id = host+1;
      *rtnd_data_len = 0;
      
      len = read(s, (char *) &nlen, sizeof(nlen));
      if (len != sizeof(nlen)) {
	PERROR ("read");
	FPRINTF (stderr, "(Net_Poll) len=%d, sizeof(nlen)=%d\n",
		 len, sizeof(nlen));
	Net_Disconnect (host+1);
	hostP->host_failed = FALSE; /* return failed only once */
	return (NET_PROCESS_FAILED);
      }
      
      nlen = ntohl(nlen);
      if (nlen > NET_MAX_PACKET_SIZE) {
	fprintf (stderr, "Net_Receive: Invalid message length %lu\n", nlen);
	return (NET_IO_ERROR);
      }
      
      toread = nlen;
      dataP = (char *) buffer_ptr;
      dataLen = MIN (nlen, NET_MAX_PACKET_SIZE);
      while (toread) {
	hasread = read (s, dataP, dataLen);
	if (hasread <= 0) {
	  PERROR ("read");
	  Net_Disconnect (host+1);
	  return (NET_IO_ERROR);
	}
	toread -= hasread;
	dataP += hasread;
	dataLen -= hasread;
      }
      
      if (net_trace_flag) {
	u_long client = hostP->host_ip_addr;
	printf ("Net_Recv: Received %d bytes from "
		"conn-id %d, host=\"%s\" [%d,%d,%d,%d] port=%d\n",
		nlen, s,  hostP->host_name, 
		(client>>24)&0xff,
		(client>>16)&0xff,(client>>8)&0xff,client&0xff,
		hostP->host_port_id);
	Net_Print_Buffer (buffer_ptr, nlen);
      }
      *rtnd_data_len = nlen;

      if (net_pause_flag) {
	char user[256];
	u_long client = hostP->host_ip_addr;
	printf ("Net_Recv: Press <ENTER> to receive "
		"data. Enter 'd' to discard.\n");
	gets(user);
	if ((strlen(user) > 0) && ((user[0] == 'd') || (user[0] == 'D')))
	  continue;
      }

      return (NET_SUCCESS);
    }
  }

  /*
   * Nothing left pending from previous read.
   * Go back and call select again
   */
  net_fd_read_pending = 0;
  goto CallSelect;
}



int Net_Receive
   (
    void *buffer_ptr,		/* ptr to buffer for incoming packet 
				 * MUST be at least MAX_PACKET_SIZE long */
    int  *rtnd_data_len,	/* returned length of packet received */
    int  *rtnd_conn_id		/* returned connection id of sender */
   )
{
  int    rc;

  rc = Net_Poll (-1, buffer_ptr, rtnd_data_len, rtnd_conn_id);
  return (rc);
}



int Net_Trace
  (
   void
  )
{
  if (net_trace_flag)
    net_trace_flag = 0;
  else
    net_trace_flag = 1;

  return (net_trace_flag);
}


int Net_Pause
  (
   void
  )
{
  if (net_pause_flag)
    net_pause_flag = 0;
  else
    net_pause_flag = 1;

  return (net_pause_flag);
}


void Net_Debug
  (
   void
  )
{
  int i;
  u_long addr;
  host_table_t *hostP;

  for (i = 0; i < net_host_entries; i++) {
    hostP = &net_hosts[i];
    addr = hostP->host_ip_addr;
    printf ("Net connection %d: to %s host \"%s\" "
	    "addr=[%d,%d,%d,%d] port=%d socket=%d%s\n",
	    i+1, 
	    (hostP->host_socket_id != NULL_SOCKET_ID) ? "running" : "FAILED",
	    hostP->host_name, 
	    (addr>>24)&0xff,(addr>>16)&0xff,(addr>>8)&0xff,addr&0xff,
	    hostP->host_port_id, hostP->host_socket_id, 
	    (hostP->host_failed) ? " Restarted" : "");
  }
}
