/* This file defines the list of tasks that a robot can perform.
   To maintain task orthogonality, each task will take the form:
   
   int (*func) (struct robot_args *arg) {};
   
   The function return and argument types may be changed later, but
   must be consistent across all tasks.
   */

#include <stdio.h>
#include <X11/Xlib.h>

#include "general.h"
#include "scheduler.h"

extern int robot_delay;
extern int sound;
extern char *sector;
extern Display *Disp;
extern void shutdown (int exit_num);

/* Currently number of degrees specifies a multiple number of 90 degrees.
   The rotational granularity may be reduced to 45 degrees to increase
   the robots degrees of freedom in the future.
   */

static task *new_task;
static struct robot_args *args;
 

int rotate_clockwise (struct robot_args *arg)
{
  robot_front[id] [arg->robot_num] =
    (robot_front[id] [arg->robot_num] + arg->value) % FULL_ROTATION;

  wipeObject (sector, robot_position[id][arg->robot_num].x, 
	      robot_position[id][arg->robot_num].y);
  if (world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y] == FRIENDLY_ROBOT)
    drawRobot (sector, robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_front[id][arg->robot_num], 
	       FRIENDLY_ROBOT, NULL);
  else
    drawRobot (sector, robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_front[id][arg->robot_num] , 
	       ENEMY_ROBOT, NULL);

#ifdef SOUND
#ifdef SOUNDCARD
  if (sound)
    playSoundFile ("rotate");
#endif
#endif

  return (T_SUCCESS);
}


int rotate_anticlockwise (struct robot_args *arg) {
  robot_front[id] [arg->robot_num] =
    (robot_front[id] [arg->robot_num] + 
     FULL_ROTATION - arg->value) % FULL_ROTATION;

  wipeObject (sector, robot_position[id][arg->robot_num].x, 
	      robot_position[id][arg->robot_num].y);

  if (world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y] == FRIENDLY_ROBOT)

    drawRobot (sector, robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_front[id][arg->robot_num], 
	       FRIENDLY_ROBOT, NULL);

  else

    drawRobot (sector, robot_position[id][arg->robot_num].x, 
	       robot_position[id][arg->robot_num].y,
	       robot_position[id][arg->robot_num].x,
	       robot_position[id][arg->robot_num].y,
	       robot_front[id][arg->robot_num] , 
	       ENEMY_ROBOT, NULL);

#ifdef SOUND
#ifdef SOUNDCARD
  if (sound)
    playSoundFile ("rotate");
#endif
#endif

  return (T_SUCCESS);
}


int checkObstacles (int location, struct robot_args *arg) {
  int bomb_underneath = FALSE;

  if ((location == POS_OBS) && 
      (robot_life[id][arg->robot_num] < INIT_LIFE*2))  {
    
    robot_life[id][arg->robot_num]++;

#ifdef SOUND
#ifdef SOUNDCARD
    if (sound)
      playSoundFile ("bonus");
#endif
#endif
  }

  else if (location == NEG_OBS) {
    robot_life[id][arg->robot_num] -= 2;
    
    if (robot_life[id][arg->robot_num] <= 0) {
      world[robot_position[id][arg->robot_num].x]
	[robot_position[id][arg->robot_num].y] = NEG_OBS;
      wipeObject (sector, 
		  robot_position[id][arg->robot_num].x, 
		  robot_position[id][arg->robot_num].y);
      drawNegObs (sector, 
		  robot_position[id][arg->robot_num].x, 
		  robot_position[id][arg->robot_num].y);
      
    }
    else
      bomb_underneath = TRUE;
    
#ifdef SOUND
#ifdef SOUNDCARD
    if (sound)
      playSoundFile ("bomb");
#endif
#endif
  }
  return (bomb_underneath);
}


void checkBombs (coordinate old_pos, struct robot_args *arg) {

  if (!killed_neg [id][arg->robot_num]) {
    world[old_pos.x][old_pos.y] = UNCHARTED;
    wipeObject (sector, old_pos.x, old_pos.y);
  }

  else {
    world[old_pos.x][old_pos.y] = NEG_OBS;
    killed_neg [id][arg->robot_num] = FALSE;
    wipeObject (sector, old_pos.x, old_pos.y);
    drawNegObs (sector, old_pos.x, old_pos.y);
  }
}  


int walk (struct robot_args *arg) {
  int tmp, tmp2;
  coordinate check;
  coordinate old_pos;
  int bomb_underneath;

  check   = robot_position[id][arg->robot_num];
  old_pos = robot_position[id][arg->robot_num];

  switch (arg->value) {
  case NORTH:
    tmp = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y];
    
    tmp2 = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y-1];

    check.y--;
    if (!in_world(check)) {

#ifdef SOUND
#ifdef SOUNDCARD
      if (sound)
	playSoundFile ("EOL");
#endif
#endif
      return (T_FAIL);
    }
    
    if ((tmp2 != FRIENDLY_ROBOT) && 
	(tmp2 != ENEMY_ROBOT) &&
	(tmp2 != STATIC_OBS) &&
	(tmp2 != UNKNOWN)) {
      (robot_position[id][arg->robot_num].y)--;

      world[robot_position[id][arg->robot_num].x]
	[robot_position[id][arg->robot_num].y] = tmp;

      bomb_underneath = checkObstacles (tmp2, arg);

      if (robot_life[id][arg->robot_num] > 0) {
	if (tmp == FRIENDLY_ROBOT)
	  drawRobot (sector, robot_position[id][arg->robot_num].x, 
		     (robot_position[id][arg->robot_num].y+1),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y, 
		     robot_front[id][arg->robot_num], 
		     FRIENDLY_ROBOT, NULL);
	else
	  drawRobot (sector, robot_position[id][arg->robot_num].x, 
		     (robot_position[id][arg->robot_num].y+1),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y,
		     robot_front[id][arg->robot_num], 
		     ENEMY_ROBOT, NULL);
      }
      checkBombs (old_pos, arg);
      if (bomb_underneath)
	killed_neg [id][arg->robot_num] = TRUE;
    }
    else
      return (T_FAIL);
    
    break;
    
  case EAST:
    tmp = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y];
    
    tmp2 = world[robot_position[id][arg->robot_num].x+1]
      [robot_position[id][arg->robot_num].y];

    check.x++;
    if (!in_world(check)) {

#ifdef SOUND
#ifdef SOUNDCARD
      if (sound)
	playSoundFile ("EOL");
#endif
#endif
      return (T_FAIL);
    }
    
    if ((tmp2 != FRIENDLY_ROBOT) && 
	(tmp2 != ENEMY_ROBOT) &&
	(tmp2 != STATIC_OBS) &&
	(tmp2 != UNKNOWN)) {
      (robot_position[id][arg->robot_num].x)++;

      world[robot_position[id][arg->robot_num].x]
	[robot_position[id][arg->robot_num].y] = tmp;

      bomb_underneath = checkObstacles (tmp2, arg);

      if (robot_life[id][arg->robot_num] > 0) {
	if (tmp == FRIENDLY_ROBOT)
	  drawRobot (sector, robot_position[id][arg->robot_num].x-1, 
		     (robot_position[id][arg->robot_num].y),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y, 
		     robot_front[id][arg->robot_num], 
		     FRIENDLY_ROBOT, NULL);
	else
	  drawRobot (sector, robot_position[id][arg->robot_num].x-1, 
		     (robot_position[id][arg->robot_num].y),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y,
		     robot_front[id][arg->robot_num], 
		     ENEMY_ROBOT, NULL);
      }
      checkBombs (old_pos, arg);
      if (bomb_underneath)
	killed_neg [id][arg->robot_num] = TRUE;
    }
    else
      return (T_FAIL);
    
    break;
    

  case SOUTH:
    tmp = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y];
    
    tmp2 = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y+1];

    check.y++;
    if (!in_world(check)) {

#ifdef SOUND
#ifdef SOUNDCARD
      if (sound)
	playSoundFile ("EOL");
#endif
#endif
      return (T_FAIL);
    }
    
    if ((tmp2 != FRIENDLY_ROBOT) && 
	(tmp2 != ENEMY_ROBOT) &&
	(tmp2 != STATIC_OBS) &&
	(tmp2 != UNKNOWN)) {
      (robot_position[id][arg->robot_num].y)++;

      world[robot_position[id][arg->robot_num].x]
	[robot_position[id][arg->robot_num].y] = tmp;

      bomb_underneath = checkObstacles (tmp2, arg);

      if (robot_life[id][arg->robot_num] > 0) {
	if (tmp == FRIENDLY_ROBOT)
	  drawRobot (sector, robot_position[id][arg->robot_num].x, 
		     (robot_position[id][arg->robot_num].y-1),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y, 
		     robot_front[id][arg->robot_num], 
		     FRIENDLY_ROBOT, NULL);
	else
	  drawRobot (sector, robot_position[id][arg->robot_num].x, 
		     (robot_position[id][arg->robot_num].y-1),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y,
		     robot_front[id][arg->robot_num], 
		     ENEMY_ROBOT, NULL);
      }
      checkBombs (old_pos, arg);
      if (bomb_underneath)
	killed_neg [id][arg->robot_num] = TRUE;
    }
    else
      return (T_FAIL);
    
    break;
    

  case WEST:
    tmp = world[robot_position[id][arg->robot_num].x]
      [robot_position[id][arg->robot_num].y];
    
    tmp2 = world[robot_position[id][arg->robot_num].x-1]
      [robot_position[id][arg->robot_num].y];

    check.x--;
    if (!in_world(check)) {

#ifdef SOUND
#ifdef SOUNDCARD
      if (sound)
	playSoundFile ("EOL");
#endif
#endif
      return (T_FAIL);
    }
    
    if ((tmp2 != FRIENDLY_ROBOT) && 
	(tmp2 != ENEMY_ROBOT) &&
	(tmp2 != STATIC_OBS) &&
	(tmp2 != UNKNOWN)) {
      (robot_position[id][arg->robot_num].x)--;

      world[robot_position[id][arg->robot_num].x]
	[robot_position[id][arg->robot_num].y] = tmp;

      bomb_underneath = checkObstacles (tmp2, arg);

      if (robot_life[id][arg->robot_num] > 0) {
	if (tmp == FRIENDLY_ROBOT)
	  drawRobot (sector, robot_position[id][arg->robot_num].x+1, 
		     (robot_position[id][arg->robot_num].y),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y, 
		     robot_front[id][arg->robot_num], 
		     FRIENDLY_ROBOT, NULL);
	else
	  drawRobot (sector, robot_position[id][arg->robot_num].x+1, 
		     (robot_position[id][arg->robot_num].y),
		     robot_position[id][arg->robot_num].x, 
		     robot_position[id][arg->robot_num].y,
		     robot_front[id][arg->robot_num], 
		     ENEMY_ROBOT, NULL);
      }
      checkBombs (old_pos, arg);
      if (bomb_underneath)
	killed_neg [id][arg->robot_num] = TRUE;
    }
    else
      return (T_FAIL);
    
    break;
    
  default :
    return (T_FAIL);
  }

#ifdef SOUND
#ifdef SOUNDCARD
  if (sound)
    playSoundFile ("walk");
#endif
#endif

  return (T_SUCCESS);
}


int initiate_missile (coordinate position, int orientation) {
  int i;

  for (i=0; missile[i].status==MISSILE_ACTIVE; i++)
    if (i == MAX_MISSILES) {
      fprintf (stderr, "Too many active missiles in system!!\n");
      shutdown(1);
    }

  missile[i].status = MISSILE_ACTIVE;
  missile[i].orientation = orientation;

  switch (orientation) {
  case NORTH:
    position.y--;
    break;

  case EAST:
    position.x++;
    break;

  case SOUTH:
    position.y++;
    break;

  case WEST:
    position.x--;
    break;

  default:
    fprintf (stderr, "missile orientation error\a\n");
    shutdown(1);
  };
  
  missile[i].pos = position;
  return i;
}


int deactivate_missile (int index) {
  missile[index].status = MISSILE_INACTIVE;
  return TRUE;
}


int get_missile_index (coordinate position) {
  
  int i;
  
  for (i = 0; (i < MAX_MISSILES) && 
	 !(equal(position, missile[i].pos)); i++);
  
  if (i == MAX_MISSILES) {
    fprintf (stderr, 
	     "Consistency error between "
	     "world and missile positions\n");
    shutdown(1);
  }
  
  return i;
}


int collision (int missile_index, coordinate position) {
  int i;
  int team_number;
  int robot_number;
  int hit = FALSE;
  
  if (!in_world (position))
    return deactivate_missile (missile_index);
  
  switch (world [position.x][position.y])
    {
    case N_MISSILE:
    case E_MISSILE:
    case S_MISSILE:
    case W_MISSILE:
      if (missile [missile_index].orientation != 
	  world [position.x][position.y]) { 
	/* We have more than one missile - collision! */

	deactivate_missile (get_missile_index (position));
	deactivate_missile (missile_index);
	world [position.x][position.y] = UNCHARTED;

	doExplosion (sector, position.x, position.y); 
	wipeObject (sector, position.x, position.y);

	hit = TRUE;
      }
      break;
      
    case FRIENDLY_ROBOT:
    case ENEMY_ROBOT:
    case UNKNOWN:
      getRobotInfo (position.x, position.y, 
		    &team_number, &robot_number);
      if (world [position.x][position.y] == UNKNOWN)
	drawUnknown (sector, position.x, position.y);
      else
	drawRobot (sector, position.x, position.y, 
		   position.x, position.y, 
		   robot_front [team_number][robot_number], 
		   world[position.x][position.y], NULL);

      robot_life [team_number][robot_number] -= MISSILE_DESTRUCTION;
      if (robot_life [team_number][robot_number] <= 0)
	world[position.x][position.y] = UNCHARTED;

      doExplosion (sector, position.x, position.y); 

      if (robot_life [team_number][robot_number] <= 0) {
	robot_life [team_number][robot_number] = 0;
	wipeObject (sector, position.x, position.y);
      }
      
      deactivate_missile (missile_index);
      hit = TRUE;
      
      break;
      
    case POS_OBS:
    case NEG_OBS:
    case STATIC_OBS:
      world[position.x][position.y] = UNCHARTED;

      doExplosion (sector, position.x, position.y); 
      wipeObject (sector, position.x, position.y);

      deactivate_missile (missile_index);
      hit = TRUE;
      break;
      
    default:
      break;
    }

  if (hit) {
      
#ifdef SOUND
#ifdef SOUNDCARD
    if (sound)
      playSoundFile ("explosion");
#endif
#endif
  }
  
  return (hit);
}


void update_missile (int missile_index) {
  int i;
  coordinate tmp;
  int redraw_goal=FALSE;

  tmp = missile[missile_index].pos;
  drawMissile (sector, missile[missile_index].pos.x, 
	       missile[missile_index].pos.y, 
	       missile[missile_index].pos.x, 
	       missile[missile_index].pos.y, 
	       missile[missile_index].orientation);
  
  switch (missile[missile_index].orientation)
    {
    case NORTH:
      for (i = 1; i < MISSILE_VELOCITY; i++)
	{
	  XFlush (Disp);
	  snooze ((int)(robot_delay*1000)/MISSILE_VELOCITY);  
	  if (collision(missile_index, missile[missile_index].pos))
	    return;
	  if (equal (missile[missile_index].pos, goal))
	    redraw_goal = TRUE;
	  missile[missile_index].pos.y--;
	  drawMissile (sector, missile[missile_index].pos.x, 
		       (missile[missile_index].pos.y+1), 
		       missile[missile_index].pos.x, 
		       missile[missile_index].pos.y, 
		       missile[missile_index].orientation);
	  if (redraw_goal)
	    redrawGoal (sector, goal);
	  redraw_goal = FALSE;
	}
      XFlush (Disp);
      if (equal (missile[missile_index].pos, goal))
	redrawGoal (sector, goal);
      if (collision(missile_index, missile[missile_index].pos))
	return;
      world[tmp.x][tmp.y] = UNCHARTED;
      world[missile[missile_index].pos.x]
	[missile[missile_index].pos.y] = N_MISSILE;
      break;
      
    case EAST:
      for (i = 1; i < MISSILE_VELOCITY; i++)
	{
	  XFlush (Disp);
	  snooze ((int)(robot_delay*1000)/MISSILE_VELOCITY);  
	  if (collision(missile_index, missile[missile_index].pos))
	    return;
	  if (equal (missile[missile_index].pos, goal))
	    redraw_goal = TRUE;
	  missile[missile_index].pos.x++;
	  drawMissile (sector, (missile[missile_index].pos.x-1), 
		       missile[missile_index].pos.y, 
		       missile[missile_index].pos.x, 
		       missile[missile_index].pos.y, 
		       missile[missile_index].orientation);
	  if (redraw_goal)
	    redrawGoal (sector, goal);
	  redraw_goal = FALSE;
	}
      XFlush (Disp);
      if (equal (missile[missile_index].pos, goal))
	redrawGoal (sector, goal);
      if (collision(missile_index, missile[missile_index].pos))
	return;
      world[tmp.x][tmp.y] = UNCHARTED;
      world[missile[missile_index].pos.x]
	[missile[missile_index].pos.y] = E_MISSILE;
      break;
      
    case SOUTH:
      for (i = 0; i < MISSILE_VELOCITY; i++)
	{
	  XFlush (Disp);
	  snooze ((int)(robot_delay*1000)/MISSILE_VELOCITY);  
	  if (collision(missile_index, missile[missile_index].pos))
	    return;
	  if (equal (missile[missile_index].pos, goal))
	    redraw_goal = TRUE;
	  missile[missile_index].pos.y++;
	  drawMissile (sector, missile[missile_index].pos.x, 
		       (missile[missile_index].pos.y-1), 
		       missile[missile_index].pos.x, 
		       missile[missile_index].pos.y, 
		       missile[missile_index].orientation);
	  if (redraw_goal)
	    redrawGoal (sector, goal);
	  redraw_goal = FALSE;
	}
      XFlush (Disp);
      if (equal (missile[missile_index].pos, goal))
	redrawGoal (sector, goal);
      if (collision(missile_index, missile[missile_index].pos))
	return;
      world[tmp.x][tmp.y] = UNCHARTED;
      world[missile[missile_index].pos.x]
	[missile[missile_index].pos.y] = S_MISSILE;
      break;
      
    case WEST:
      for (i = 0; i < MISSILE_VELOCITY; i++)
	{
	  XFlush (Disp);
	  snooze ((int)(robot_delay*1000)/MISSILE_VELOCITY);  
	  if (collision(missile_index, missile[missile_index].pos))
	    return;
	  if (equal (missile[missile_index].pos, goal))
	    redraw_goal = TRUE;
	  missile[missile_index].pos.x--;
	  drawMissile (sector, (missile[missile_index].pos.x+1), 
		       missile[missile_index].pos.y, 
		       missile[missile_index].pos.x, 
		       missile[missile_index].pos.y, 
		       missile[missile_index].orientation);
	  if (redraw_goal)
	    redrawGoal (sector, goal);
	  redraw_goal = FALSE;
	}
      XFlush (Disp);
      if (equal (missile[missile_index].pos, goal))
	redrawGoal (sector, goal);
      if (collision(missile_index, missile[missile_index].pos))
	return;
      world[tmp.x][tmp.y] = UNCHARTED;
      world[missile[missile_index].pos.x]
	[missile[missile_index].pos.y] = W_MISSILE;
      break;
      
    default:
      break;
    }
}


/* arg->value holds the team number for which the 
   robot that fired is a member. */
int fire_missile (struct robot_args *arg) {

  coordinate c;

  if (arg->orientation != -1) {
    c.x = arg->x;
    c.y = arg->y;
    
    initiate_missile (c, arg->orientation);
  }
  else
    initiate_missile (robot_position [arg->value][arg->robot_num],
		      robot_front [arg->value][arg->robot_num]);
  
#ifdef SOUND
#ifndef SOUNDCARD
  XBell (Disp, 100);
  XFlush (Disp);
#else
  if (sound)
    playSoundFile ("missile");
#endif
#endif

  return (T_SUCCESS);
}


void create_task (int sched_num,
		  int robot_id,
		  int sighted_flag[MAX_ROBOTS]
		  [NUM_OBJECT_TYPES+OBJECT_OFFSET],
		  coordinate environment 
		  [MAX_ROBOTS][NUM_OBJECT_TYPES+OBJECT_OFFSET])
{
  int i, walk_direction;
  int direction;
  
  walk_direction = calculate_walk_orientation(robot_id);
  if ((walk_direction < NORTH)||(walk_direction>WEST))
    {
      fprintf (stderr, 
	       "Processing robot %d\tError in Walk direction %d\n", 
	       robot_id, walk_direction);
      shutdown(1);
    }
  
#if 0 /* ?? Test*/
  if (sighted_flag [robot_id][MISSILE]) {
      
      if (abs (robot_front[id] [robot_id] - 
	       world [environment [robot_id][MISSILE].x]
	       [environment [robot_id][MISSILE].y]) == 2) 
	/* Robot is facing incoming missile. */
	{
	  /* Creating fire missile task to destroy approaching missile. */
	  new_task = (task *) malloc (sizeof (task));
	  args     = (struct robot_args *) 
	    malloc (sizeof (struct robot_args));
	  
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1;
	  new_task->deadline         = current_time + 1;
	  new_task->initial_priority = 8;
	  new_task->value_func       = STEP_FUNCTION;
	  new_task->func             = fire_missile;
	  
	  args ->robot_num  = robot_id;
	  args ->value      = id;
	  args->orientation = -1;
	  
	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
      else
	{
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
	  
	  if (robot_front[id] [robot_id] - 
	      world [environment [robot_id][MISSILE].x]
	      [environment [robot_id][MISSILE].y] > 0)
	    {
	      new_task->func  = rotate_anticlockwise;
	      args ->value    = 
		(robot_front[id] [robot_id] + FULL_ROTATION -
		 world [environment [robot_id][MISSILE].x]
		 [environment [robot_id][MISSILE].y]) % FULL_ROTATION;
	    }
	  
	  else
	    if ((robot_front[id] [robot_id] - 
		 world [environment [robot_id][MISSILE].x]
		 [environment [robot_id][MISSILE].y] < 0) &&
		already_visited (environment [robot_id][MISSILE])) 
	      {
		new_task->func  = rotate_clockwise;
		args ->value    = 
		  (world [environment [robot_id][MISSILE].x]
		   [environment [robot_id][MISSILE].y] -
		   robot_front[id] [robot_id]) % FULL_ROTATION;
	      }

	  /* Creating a rotation task. 
	     Prepare to destroy approaching missile. */
	  
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1; /*args ->value;*/
	  new_task->deadline         = current_time + 1; /*args->value;*/
	  new_task->initial_priority = 6;
	  new_task->value_func       = RAMP_FUNCTION;
	  
	  args ->robot_num = robot_id;
	  
	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);

	  /* Creating a walk task to avoid enemy missile. */
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
      
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1;
	  new_task->deadline         = current_time + 1;
	  new_task->initial_priority = 10; 
	  new_task->value_func       = STEP_FUNCTION;
	  new_task->func             = walk;
	  
	  args ->robot_num = robot_id;
	  args ->value     = robot_front[id][robot_id];
      
	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
    }
  
  if (sighted_flag [robot_id][ENEMY_ROBOT])
    {
      direction = next_move (robot_position[id][robot_id], 
			     environment [robot_id][ENEMY_ROBOT]);
      
      if (robot_front[id] [robot_id] == direction)
	/* Robot is facing enemy robot. */
	{
	  /* Creating a fire missile task against enemy robot. */
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
	  
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1;
	  new_task->deadline         = current_time + 1;
	  
	  new_task->initial_priority = 9; /* used to be 8 */
	  new_task->value_func       = STEP_FUNCTION; /* used to be ramp */
	  new_task->func             = fire_missile;
	  
	  args ->robot_num  = robot_id;
	  args ->value      = id;
	  args->orientation = -1;

	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
      else
	{
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
	  
	  direction = next_move (robot_position[id][robot_id], 
				 environment [robot_id][ENEMY_ROBOT]);

	  if (robot_front[id] [robot_id] - direction > 0)
	    {
	      new_task->func  = rotate_anticlockwise;
	      args ->value     = 
		(robot_front[id] [robot_id] + FULL_ROTATION -
		 direction) % FULL_ROTATION;
	    }
	  else
	    if (robot_front[id] [robot_id] - direction < 0)
	      {
		new_task->func  = rotate_clockwise;
		args ->value     = 
		  (direction - robot_front[id] [robot_id]) % FULL_ROTATION;
	      }
	  
	  /* Creating a rotation task. Prepare to fire upon detected enemy. */
	    
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1; /*args ->value;*/
	  new_task->deadline         = current_time + 1; /*args->value;*/
	  new_task->initial_priority = 6;
	  new_task->value_func       = RAMP_FUNCTION;
	  
	  args ->robot_num = robot_id;
	  
	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
    }

  if (sighted_flag [robot_id][NEG_OBS]) 
    {
      direction = next_move (robot_position[id][robot_id], 
			     environment [robot_id][NEG_OBS]); 
      
      if (robot_front[id] [robot_id] == direction)
	/* Robot is facing bomb. */
	{
	  /* Creating a fire missile task to destroy sighted bomb. */
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
	  
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1;
	  new_task->deadline         = current_time + 1;
	  new_task->initial_priority = 2;
	  new_task->value_func       = RAMP_FUNCTION;
	  new_task->func             = fire_missile;
	  
	  args ->robot_num  = robot_id;
	  args ->value      = id;
	  args->orientation = -1;
	  
	  new_task->arg = args ;
	  
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
      else
	{
	  new_task = (task *) malloc (sizeof (task));
	  args     = 
	    (struct robot_args *) malloc (sizeof (struct robot_args));
	  
	  direction = next_move (robot_position[id][robot_id], 
				 environment [robot_id][NEG_OBS]); 

	  if (robot_front[id] [robot_id] - direction > 0)
	    {
	      new_task->func  = rotate_anticlockwise;
	      args ->value     = 
		(robot_front[id] [robot_id] + FULL_ROTATION -
		 direction) % FULL_ROTATION;
	    }
	  else
	    if (robot_front[id] [robot_id] - direction < 0)
	      {
		new_task->func  = rotate_clockwise;
		args ->value     = 
		  (direction - robot_front[id] [robot_id]) % FULL_ROTATION;
	      }
	      
	  /* Creating a rotation task. Prepare to fire upon nearby bomb. */
	    
	  new_task->arrival_time     = current_time;
	  new_task->start_time       = current_time;
	  new_task->run_time         = 1; /*args ->value;*/
	  new_task->deadline         = current_time + 1; /*args->value;*/
	  new_task->initial_priority = 2;
	  new_task->value_func       = RAMP_FUNCTION;
	  
	  args ->robot_num = robot_id;
	  
	  new_task->arg = args ;
	    
	  add_to_buffer (new_task, sched_num,
			 buffer [sched_num].front,
			 &buffer [sched_num].back,
			 &buffer [sched_num].status);
	}
    }
#endif /* ?? Test*/
  
  if (sighted_flag [robot_id][POS_OBS])
    {
      direction = next_move (robot_position[id][robot_id], 
			     environment [robot_id][POS_OBS]);

      if (robot_front[id] [robot_id] == direction) { 
	/* Robot is facing food. */
	/* Creating a walk task. Food sighted. */
	new_task = (task *) malloc (sizeof (task));
	args     = 
	  (struct robot_args *) malloc (sizeof (struct robot_args));
      
	new_task->arrival_time     = current_time;
	new_task->start_time       = current_time;
	new_task->run_time         = 1;
	new_task->deadline         = current_time + 1;
	new_task->initial_priority = 4;
	new_task->value_func       = RAMP_FUNCTION;
	new_task->func             = walk;
      
	args ->robot_num = robot_id;
	args ->value     = robot_front[id][robot_id];
	
	new_task->arg = args ;
	
	add_to_buffer (new_task, sched_num,
		       buffer [sched_num].front,
		       &buffer [sched_num].back,
		       &buffer [sched_num].status);
      }
    }
  
  if (sighted_flag [robot_id][STATIC_OBS]) {
    direction = next_move (robot_position[id][robot_id], 
			   environment [robot_id][STATIC_OBS]); 
    
    if (robot_front[id] [robot_id] == direction) {
      /* Robot is facing obstacle. */ 
      
      /* Creating a fire missile task to destroy sighted obstacle. */
      new_task = (task *) malloc (sizeof (task));
      args     = 
	(struct robot_args *) malloc (sizeof (struct robot_args));
      
      new_task->arrival_time     = current_time;
      new_task->start_time       = current_time;
      new_task->run_time         = 1;
      new_task->deadline         = current_time + 1;
      new_task->initial_priority = 3;
      new_task->value_func       = RAMP_FUNCTION;
      new_task->func             = fire_missile;
      
      args ->robot_num  = robot_id;
      args ->value      = id;
      args->orientation = -1;
	  
      new_task->arg = args ;
      
      add_to_buffer (new_task, sched_num,
		     buffer [sched_num].front,
		     &buffer [sched_num].back,
		     &buffer [sched_num].status);
    }
    else
      {
	new_task = (task *) malloc (sizeof (task));
	args     = 
	  (struct robot_args *) malloc (sizeof (struct robot_args));
	
	direction = next_move (robot_position[id][robot_id], 
			       environment [robot_id][STATIC_OBS]); 
	
	if (robot_front[id] [robot_id] - direction > 0)
	  {
	    new_task->func  = rotate_anticlockwise;
	    args ->value     = 
	      (robot_front[id] [robot_id] + FULL_ROTATION -
	       direction) % FULL_ROTATION;
	  }
	else
	  if (robot_front[id] [robot_id] - direction < 0)
	    {
	      new_task->func  = rotate_clockwise;
	      args ->value     = 
		(direction - robot_front[id] [robot_id]) % FULL_ROTATION;
	    }
	
	/* Creating a rotation task. Prepare to fire upon nearby obstacle. */
	
	new_task->arrival_time     = current_time;
	new_task->start_time       = current_time;
	new_task->run_time         = 1; /*args ->value;*/
	new_task->deadline         = current_time + 1; /*args->value;*/
	new_task->initial_priority = 2;
	new_task->value_func       = RAMP_FUNCTION;
	
	args ->robot_num = robot_id;
	
	new_task->arg = args ;
	
	add_to_buffer (new_task, sched_num,
		       buffer [sched_num].front,
		       &buffer [sched_num].back,
		       &buffer [sched_num].status);
      }
  }

  if (robot_front[id][robot_id] == walk_direction) {
    /* Creating a walk task. Let's walk to the goal. */
    new_task = (task *) malloc (sizeof (task));
    args     = (struct robot_args *) malloc (sizeof (struct robot_args));
    
    new_task->arrival_time     = current_time;
    new_task->start_time       = current_time;
    new_task->run_time         = 1;
    new_task->deadline         = current_time + 1 /*NO_DEADLINE*/;
    new_task->initial_priority = 2;
    new_task->value_func       = RAMP_FUNCTION;
    new_task->func             = walk;
    
    args ->robot_num = robot_id;
    args ->value     = robot_front[id][robot_id];
    
    new_task->arg = args ;
    
    add_to_buffer (new_task, sched_num,
		   buffer [sched_num].front,
		   &buffer [sched_num].back,
		   &buffer [sched_num].status);
  }
  else
    {
      /* Creating a rotate task. 
	 Let's turn towards the goal. */
      new_task = (task *) malloc (sizeof (task));
      args     = (struct robot_args *) malloc (sizeof (struct robot_args));
      
      if (robot_front[id] [robot_id] - walk_direction > 0)
	{
	  new_task->func  = rotate_anticlockwise;
	  args ->value     = (robot_front[id] [robot_id] + FULL_ROTATION -
			      walk_direction) % FULL_ROTATION;
	}
      else
	if (robot_front[id] [robot_id] - walk_direction < 0)
	  {
	    new_task->func  = rotate_clockwise;
	    args ->value    = 
	      (walk_direction - robot_front[id] [robot_id]) % 
	      FULL_ROTATION;
	  }
      
      new_task->arrival_time     = current_time;
      new_task->start_time       = current_time;
      new_task->run_time         = 1; /*args ->value;*/
      new_task->deadline         = current_time + 1; /*args->value;*/
      new_task->initial_priority = 2;
      new_task->value_func       = RAMP_FUNCTION;
      
      args ->robot_num = robot_id;
      
      new_task->arg = args ;
      
      add_to_buffer (new_task, sched_num,
		     buffer [sched_num].front,
		     &buffer [sched_num].back,
		     &buffer [sched_num].status);
    }
}


/* User-mode is in operation. 
   Tasks are created according to user's mouse events. */

void createTaskFromInput (int sched_num, 
			  int robot_id, 
			  int button_number,
			  int rotation) {

  if ((robot_life [id][robot_id] > 0) && 
      (!equal (robot_position[id] [robot_id], goal))) {
    new_task = (task *) malloc (sizeof (task));
    args     = 
      (struct robot_args *) malloc (sizeof (struct robot_args));
    
    new_task->arrival_time     = current_time;
    new_task->start_time       = current_time;
    new_task->run_time         = 1;
    new_task->deadline         = current_time + 1;
    new_task->initial_priority = 10; 
    new_task->value_func       = STEP_FUNCTION;
    
    if (button_number == 1) {
      /* Create a walk task in direction of robot_front[id]. */
      args ->value     = robot_front[id][robot_id];
      new_task->func = walk;
    }
    else
      if (button_number == 2) {
	/* Create a rotate task. */
	args ->value     = 1;
	if (rotation == CLOCKWISE)
	  new_task->func = rotate_clockwise;
	else
	  new_task->func = rotate_anticlockwise;
      }
      else
	if (button_number == 3) {/* Create a fire missile task. */
	  args ->value     = id;
	  new_task->func = fire_missile;
	}
    
    args ->robot_num   = robot_id;
    args ->orientation = -1;
    new_task->arg      = args ;
    
    add_to_buffer (new_task, sched_num,
		   buffer [sched_num].front,
		   &buffer [sched_num].back,
		   &buffer [sched_num].status);
  }
}
